#include <stdio.h>
#include <stdlib.h>
#include <signal.h>

#include <X11/Intrinsic.h>
#include <X11/StringDefs.h>
#include <X11/keysym.h>

#include "dftype.h"
#include "dfxfunc.h"

static int CursorUp(Widget w, DfFileList *list);
static int CursorDown(Widget w, DfFileList *list);
static int PageUp(Widget w, DfFileList *list);
static int PageDown(Widget w, DfFileList *list);
static void DoOpen(Widget w, DfFileList *list);
static void *NewBuffer(DfFileList *list);
static void SwitchBuffer(Widget w, DfFileList *list);
static int IsSingleBuffer(void);
static DfBuf *SplitWindowHr(Widget w, DfFileList *list);
static DfBuf *ChooseBufferSplit(DfBuf *buf);

static void KillBuffer(Widget w, DfFileList *list);

static int ChDir(DfFileList *list, const char *new_dir);
static void SelectFile(Widget w, DfFileList *list);
static int GotoParent(Widget w, DfFileList *list);

static void CalcCursorRect(DfFileList *list, int nPos, XRectangle *rc);
extern int ShowDialog(Widget w);

static int FL_Free(void *list);
static int FL_Active(void *ptr);
static int FL_KeyPress(void *ptr, Widget w, KeySym k, Boolean *cond);
static int FL_Draw(void *ptr, Display *display, Widget w, GC gc);


static const DfVerb FL_Verb = {
  FL_Active,
  FL_KeyPress,
  FL_Draw,
  FL_Free,
};

enum{
  NOP = 0,
  OPEN,
  CURUP,
  CURDOWN,
  PAGEUP,
  PAGEDOWN,
  SELECT,
  SPRIT_H,
  SPRIT_V,
  CLOSE,
  NEWBUFFER,
  SWITCHBUFFER,
  GOTOPARENT,
  COPY,
  QUIT,
};

DfCmdKey c_strOpen[] ={
  {          0, XK_Return},
  {          0, 0}
};
DfCmdKey c_strUp[] ={
  {          0, XK_Up},
  {          0, 0}
};
DfCmdKey c_strDown[] ={
  {          0, XK_Down},
  {          0, 0}
};
DfCmdKey c_strPageUp[] ={
  {          0, XK_Prior},
  {          0, 0}
};
DfCmdKey c_strPageDown[] ={
  {          0, XK_Next},
  {          0, 0}
};
DfCmdKey c_strSelect[] ={
  {          0, XK_space},
  {          0, 0}
};

DfCmdKey c_strSpritH[] ={
  {          0, XK_w},
  {          0, XK_w},
  {          0, 0}
};

DfCmdKey c_strSpritV[] ={
  {          0, XK_w},
  {          0, XK_v},
  {          0, 0}
};

DfCmdKey c_strClose[] ={
  {          0, XK_w},
  {          0, XK_q},
  {          0, 0}
};

DfCmdKey c_strNewBuf[] ={
  {          0, XK_n},
  {          0, XK_e},
  {          0, XK_w},
  {          0, 0}
};

DfCmdKey c_strSwitchBuf[] ={
  {          0, XK_Tab},
  {          0, 0}
};

DfCmdKey c_strGotoParent[] ={
  {          0, XK_BackSpace},
  {          0, 0}
};


DfCmdKey c_strCopy[] ={
  {          0, XK_c},
  {          0, XK_p},
  {          0, 0}
};

DfCmdKey c_strQuit[] ={
  {          0, XK_q},
  {          0, 0}
};

DfCmdKey c_strCtrlH[] ={
  { DFMOD_CTRL, XK_h},
  {           0, 0}
};


static const DfCmdStr FL_CmdTab[] = {
  {c_strOpen, OPEN},
  {c_strUp, CURUP},
  {c_strDown, CURDOWN},
  {c_strPageUp, PAGEUP},
  {c_strPageDown, PAGEDOWN},
  {c_strSelect, SELECT},
  {c_strSpritH, SPRIT_H},
  {c_strSpritV, SPRIT_V},
  {c_strClose, CLOSE},
  {c_strNewBuf, NEWBUFFER},
  {c_strSwitchBuf, SWITCHBUFFER},
  {c_strGotoParent, GOTOPARENT},
  {c_strCtrlH, GOTOPARENT},
  {c_strCopy, COPY},
  {c_strQuit, QUIT},
  {NULL, 0},
};

static int DoCopy(Widget w, DfFileList  *list)
{
  DfStr s;
  int n;
  int i;
  char *p;

  n = MakeSelectList(&s, list);
  p = Str_Get(&s);
  printf("%d files select.\n", list->nSelects);
  for(i = 0; i < n; i++){
    printf("%d:%s\n", i + 1, p);
    p += (strlen(p) + 1);
  }

  Str_Free(&s, 0);

  ShowDialog(w);

  return 0;
}

/* filer */
void FL_Init(DfFileList *list)
{
  memset(list, 0, sizeof(*list));
  ListClear(list);
  list->b.v = &FL_Verb;
}

static int FL_Free(void *ptr)
{
  DfFileList *list;

  list = ptr;

  bFree(list->pszCurDir);
  bFree(list->files);
  bFree(list->pszFilenames);
  bFree(list);

  return 0;
}

static int FL_Active(void *ptr)
{
  CmdSetTable(FL_CmdTab);
  return 0;
}

static int FL_KeyPress(void *ptr, Widget w, KeySym k, Boolean *cond)
{
  DfFileList *list;
  int redraw;
  int pre_cur;
  XRectangle rc;
  DfBuf *b;
  int cmd;

  cmd = CmdTypeAhead(k);
  if(cmd == 0){
    return 0;
  }

  redraw = 0;

  list = ptr;
  pre_cur = list->nCur;
  

  switch(cmd){
  case COPY:
    DoCopy(w, list);
    break;
  case QUIT:
    if(1 < dfx->nBuf){
      KillBuffer(w, list);
    }else{
      exit(0);
    }
    return 0;
    break;
  case SPRIT_H:
    b = SplitWindowHr(w, list);
    InvalidateRect(w, &b->rc);
    redraw = 1;
    break;
  case NEWBUFFER:
    NewBuffer(list);
    SwitchBuffer(w, list);
    break;
  case XK_x:
#if MEMDEBUG
    DumpMemInfo();
#endif
    break;
  case CURUP:
    redraw = CursorUp(w, list);
    break;
  case CURDOWN:
    redraw = CursorDown(w, list);
    break;
  case SWITCHBUFFER:
    SwitchBuffer(w, list);
    break;
  case OPEN:
    DoOpen(w, list);
    break;
  case SELECT:
    SelectFile(w, list);
    redraw = CursorDown(w, list);
    if(!redraw && pre_cur == list->nCur){
      /* cursor is not moved */
      CalcCursorRect(list, list->nCur, &rc);
      XClearArea(XtDisplay(w), XtWindow(w), rc.x, rc.y, rc.width, rc.height, True);
    }
    break;
  case PAGEUP:
    redraw = PageUp(w, list);
    break;
  case PAGEDOWN:
    redraw = PageDown(w, list);
    break;
  case GOTOPARENT:
    redraw = GotoParent(w, list);
    break;
  }

  if(redraw){
    InvalidateRect(w, &list->b.rc);
    return 0;
  }
  if(pre_cur != list->nCur){
    CalcCursorRect(list, pre_cur, &rc);
    InvalidateRect(w, &rc);
    CalcCursorRect(list, list->nCur, &rc);
    InvalidateRect(w, &rc);
  }
  return 0;
}

static int FL_Draw(void *ptr, Display *display, Widget w, GC gc)
{
  DfFile *files;
  DfFile *cur;
  DfFileList *list;
  int x;
  int y;
  int cx;
  int cy;
  int n;
  int nResidual;
  long int fg;
  long int bg;
  int right;
  XRectangle rc;

  /* ToDo:
   * multi file lists.
   */
  list = ptr;
  cx = list->b.rc.width;
  cy = vnFontHeight;
  files = list->files + list->nStart;
  nResidual = list->nFiles - list->nStart;
  cur = list->files + list->nCur;
  right = list->b.rc.x + list->b.rc.width;
  for(x = list->b.rc.x; x < right; x += cx){
    n = 0;
    for(y = vnFontBase; y - vnFontBase < list->b.rc.height; y += cy){
      fg = files->color;
      bg = colors[COLOR_BK];
      if(files == cur){
	fg = colors[COLOR_CURFG];
	bg = colors[COLOR_CURBG];
      }else if(files->nSelect){
	fg = colors[COLOR_SELFG];
	bg = colors[COLOR_SELBG];
      }
      CalcCursorRect(list, n + list->nStart, &rc);
      XSetForeground(display, gc, bg);
      XFillRectangle(XtDisplay(w), XtWindow(w), gc, rc.x, rc.y, rc.width, rc.height);

      if(files->nSelect){
	XSetForeground(display, gc, colors[COLOR_MARK]);
	XFillRectangle(XtDisplay(w), XtWindow(w), gc, rc.x, rc.y, rc.x + 4, rc.height);
      }
      XSetForeground(display, gc, fg);
      XmbDrawString(display, XtWindow(w), vFontSet, gc, x + 4, y, files->pszName, strlen(files->pszName));

      files++;
      nResidual--;
      if(nResidual == 0){
	goto DRAW_FINISH;
      }
      n++;
    }
  }

DRAW_FINISH:
  return 0;
}

static int CursorUp(Widget w, DfFileList *list)
{
  int n;

  n = list->nCur;
  if(n){
    n--;
    list->nCur = n;
    if(n < list->nStart){
      list->nStart--;
      return 1;
    }
  }
 
  return 0;
}

static int CursorDown(Widget w, DfFileList *list)
{
  int n;

  n = list->nCur;
  if(n < list->nFiles - 1){
    n++;
    list->nCur = n;
    if(list->nStart + list->b.rc.height / vnFontHeight - 1 < n){
      /* scroll */
      list->nStart++;
      //     XCopyArea(XtDisplay(w), XtWindow(w), XtWindow(w), DefaultGC(XtDisplay(w), 0), 0, vnFontHeight, list->width, list->height, 0, 0);
      return 1;
    }
  }

  return 0;
}

static void DoOpen(Widget w, DfFileList *list)
{
  int cur;
  int attr;

  cur = list->nCur;
  attr = list->files[cur].attr;

  if(S_ISDIR(attr)){
    if(ChDir(list, list->files[cur].pszName)){
      InvalidateWindow(w);
    }
    return;
  }
}

static int ChDir(DfFileList *list, const char *new_dir)
{
  printf("new: %s\n", new_dir);
  if(GetFiles(list, new_dir)){
    list->nCur = 0;
    list->nStart = 0;
    return 1;
  }

  return 0;
}

static void *NewBuffer(DfFileList *list)
{
  DfFileList *new_buffer;

  new_buffer = bMalloc(sizeof(*new_buffer));
  FL_Init(new_buffer);
  new_buffer->nCur = 0;
  new_buffer->nStart = 0;
  new_buffer->pszCurDir = NULL;
  new_buffer->b.rc = list->b.rc;

  GetFiles(new_buffer, list->pszCurDir);
  ListAdd(list, new_buffer);
  dfx->nBuf++;

  return new_buffer;
}

static void KillBuffer(Widget w, DfFileList *list)
{
  DfBuf *b;

  if(dfx->nBuf <= 1){
    /* must have least 1 buffer */
    return;
  }

  b = NextBuffer(list);
  if(dfx->active == &list->b){
    dfx->active = b;
  }
  if(dfx->lists == &list->b){
    dfx->lists = b;
  }
  ListDel(list);
  dfx->nBuf--;

  if(b->visible){
    RectAdd(&b->rc, &list->b.rc);
    InvalidateRect(w, &b->rc);
  }else{
    InvalidateRect(w, &list->b.rc);
  }

  list->b.v->free(list);
}


static void SwitchBuffer(Widget w, DfFileList *list)
{
  DfBuf *b;

  list->b.visible = 0;
  b = (DfBuf*)list;
  do{
    b = NextBuffer(b);
  }while(b->visible);
  
  dfx->active = b;
  b->visible = 1;
  b->rc = list->b.rc;
  InvalidateRect(w, &b->rc);
}

static DfBuf *ChooseBufferSplit(DfBuf *buf)
{
  DfBuf *n;

  if(IsSingleBuffer()){
    n = NewBuffer((DfFileList*)buf);
  }else{
    n = NewBuffer((DfFileList*)buf);
    while(n->visible && n != buf){
      /* skip visibled buffer */
      n = NextBuffer(n);
    }
    if(n == buf){
      /* all buffers are visibled, make new buffer. */
      n = NewBuffer((DfFileList*)buf);
    }
  }

  return n;
}


static DfBuf *SplitWindowHr(Widget w, DfFileList *list)
{
  DfBuf *n;
  int width;

  n = ChooseBufferSplit((DfBuf*)list);

  width = list->b.rc.width;
  list->b.rc.width /= 2;

  n->rc.x = list->b.rc.width;
  n->rc.width = width - list->b.rc.width;
  n->visible = 1;

  return n;
}

static int IsSingleBuffer(void)
{
  return dfx->nBuf == 1;
}

static void SelectFile(Widget w, DfFileList *list)
{
  int n;
  DfFile *f;
  int i;
  XRectangle rc;

  n = list->nCur;

  if(list->nFiles <= n){
    return;
  }

  f = &list->files[n];

  if(f->nSelect){
    n = f->nSelect;
    f->nSelect = 0;
    f = list->files;
    for(i = 0; i < list->nFiles; i++){
      if(n < f->nSelect){
	f->nSelect--;
      }
      f++;
    }
    list->nSelects--;
  }else{
    list->nSelects++;
    f->nSelect = list->nSelects;
  }
  CalcCursorRect(list, n - list->nStart, &rc);
  XClearArea(XtDisplay(w), XtWindow(w), rc.x, rc.y, rc.width, rc.height, True);
}

static int GotoParent(Widget w, DfFileList *list)
{
  char *p;

  p = FindLastChar(list->pszCurDir, '/');
  if(!p){
    return 0;
  }

  if(p[1] == '\0' && p == list->pszCurDir){
    /*  /  */
    return 0;
  }

  if(p[1] == '\0'){
    p[0] = '\0';
    p = FindLastChar(list->pszCurDir, '/');
  }
  p[1] = '\0';
  return GetFiles(list, NULL);
}

static int PageUp(Widget w, DfFileList *list)
{
  int n;
  int pre;

  pre = list->nCur;

  n = list->b.rc.height / vnFontHeight;
  switch(n){/* :D  */
  case 0:
    break;
  default:
    n--;
  }
  if(list->nCur <= n){
    list->nCur = 0;
  }else{
    list->nCur -= n;
  }
  if(list->nStart <= n){
    list->nStart = 0;
  }else{
    list->nStart -= n;
  }

  if(pre != list->nCur){
    return 1;
  }
  return 0;
}

static int PageDown(Widget w, DfFileList *list)
{
  int n;
  int pre;

  pre = list->nCur;

  n = list->b.rc.height / vnFontHeight;
  switch(n){/* :D  */
  case 0:
  case 1:
    break;
  default:
    n--;
  }
  list->nCur += n;
  if(list->nFiles <= list->nCur){
    list->nCur = list->nFiles - 1;
  }

  if(list->nStart + n < list->nCur){
    list->nStart += n;
    if(list->nFiles - n - 1 <= list->nStart){
      list->nStart = list->nFiles - n - 1;
    }
  }

  if(pre != list->nCur){
    return 1;
  }
  return 0;
}

static void CalcCursorRect(DfFileList *list, int nPos, XRectangle *rc)
{
  nPos -= list->nStart;
  rc->x = list->b.rc.x;
  rc->y = list->b.rc.y + nPos * vnFontHeight;
  rc->width = list->b.rc.width;
  rc->height = vnFontHeight;
}
